"""
Google OpenIdConnect:
    https://python-social-auth.readthedocs.io/en/latest/backends/google.html
"""

from typing import Any

from .google import GoogleOAuth2
from .open_id_connect import OpenIdConnectAuth


class GoogleOpenIdConnect(GoogleOAuth2, OpenIdConnectAuth):
    name = "google-openidconnect"
    OIDC_ENDPOINT = "https://accounts.google.com"
    # differs from value in discovery document
    # http://openid.net/specs/openid-connect-core-1_0.html#rfc.section.15.6.2
    ID_TOKEN_ISSUER = "accounts.google.com"

    def get_user_id(self, details, response):
        if self.setting("USE_UNIQUE_USER_ID", False) and self.id_token is not None:
            return self.id_token.get("sub")
        return super().get_user_id(details, response)

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Return user data from Google API"""
        return self.validate_userinfo_sub(
            self.get_json(
                "https://openidconnect.googleapis.com/v1/userinfo",
                params={"access_token": access_token, "alt": "json"},
            )
        )



import json

import responses

from .oauth import BaseAuthUrlTestMixin, OAuth2StateTestMixin, OAuth2Test


class CleverOAuth2Test(OAuth2Test, OAuth2StateTestMixin, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.clever.CleverOAuth2"
    expected_username = "ada"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})

    def auth_handlers(self, start_url: str) -> str:
        target_url = super().auth_handlers(start_url)
        responses.add(
            responses.GET,
            "https://api.clever.com/v3.0/me",
            body=json.dumps({"data": {"id": "user-id"}}),
            content_type="application/json",
        )
        responses.add(
            responses.GET,
            "https://api.clever.com/v3.0/users/user-id",
            body=json.dumps(
                {
                    "data": {
                        "id": "user-id",
                        "name": {"first": "Ada", "last": "Lovelace"},
                        "email": "ada@example.com",
                        "roles": {
                            "student": {"credentials": {"district_username": "ada"}}
                        },
                    }
                }
            ),
            content_type="application/json",
        )
        return target_url

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Github Enterprise OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/github_enterprise.html
"""

from __future__ import annotations

from typing import TYPE_CHECKING, cast
from urllib.parse import urljoin

from social_core.exceptions import AuthMissingParameter
from social_core.utils import append_slash

from .github import GithubOAuth2, GithubOrganizationOAuth2, GithubTeamOAuth2

if TYPE_CHECKING:
    from .base import BaseAuth


class GithubEnterpriseMixin:
    def _required_setting(self, name: str) -> str:
        value = cast("GithubOAuth2", self).setting(name)
        if not value:
            raise AuthMissingParameter(cast("BaseAuth", self), name)
        return cast("str", value)

    def api_url(self):
        return append_slash(self._required_setting("API_URL"))

    def authorization_url(self):
        return self._url("login/oauth/authorize")

    def access_token_url(self):
        return self._url("login/oauth/access_token")

    def _url(self, path):
        return urljoin(append_slash(self._required_setting("URL")), path)


class GithubEnterpriseOAuth2(GithubEnterpriseMixin, GithubOAuth2):
    """Github Enterprise OAuth authentication backend"""

    name = "github-enterprise"


class GithubEnterpriseOrganizationOAuth2(
    GithubEnterpriseMixin, GithubOrganizationOAuth2
):
    """Github Enterprise OAuth2 authentication backend for
    organizations"""

    name = "github-enterprise-org"
    DEFAULT_SCOPE = ["read:org"]


class GithubEnterpriseTeamOAuth2(GithubEnterpriseMixin, GithubTeamOAuth2):
    """Github Enterprise OAuth2 authentication backend for teams"""

    name = "github-enterprise-team"
    DEFAULT_SCOPE = ["read:org"]



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2StateTestMixin, OAuth2Test


class GoClioOAuth2Test(OAuth2Test, OAuth2StateTestMixin, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.goclio.GoClioOAuth2"
    user_data_url = "https://app.goclio.com/api/v2/users/who_am_i"
    expected_username = "123"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "user": {
                "id": "123",
                "email": "ada@example.com",
                "first_name": "Ada",
                "last_name": "Lovelace",
            }
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()


class GoClioEuOAuth2Test(GoClioOAuth2Test):
    backend_path = "social_core.backends.goclioeu.GoClioEuOAuth2"
    user_data_url = "https://app.goclio.eu/api/v2/users/who_am_i"



from typing import Any

from .open_id_connect import OpenIdConnectAuth


class PixelPinOpenIDConnect(OpenIdConnectAuth):
    """PixelPin OpenID Connect authentication backend"""

    name = "pixelpin-openidconnect"
    ID_KEY = "sub"
    AUTHORIZATION_URL = "https://login.pixelpin.io/connect/authorize"
    ACCESS_TOKEN_URL = "https://login.pixelpin.io/connect/token"
    OIDC_ENDPOINT = "https://login.pixelpin.io"
    JWKS_URI = "https://login.pixelpin.io/.well-known/jwks"
    REQUIRES_EMAIL_VALIDATION = False

    def get_user_details(self, response):
        """Return user details from PixelPin account"""
        first_name = response.get("given_name")
        last_name = response.get("family_name")
        sub = response.get("sub")
        if sub is None and self.id_token is not None:
            sub = self.id_token.get("sub")

        username = first_name + last_name + sub

        return {
            "username": username,
            "email": response.get("email"),
            "fullname": f"{first_name} {last_name}",
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        return self.validate_userinfo_sub(
            self.get_json(
                "https://login.pixelpin.io/connect/userinfo",
                headers={"Authorization": f"Bearer {access_token}"},
            )
        )



"""
Globus Auth OpenID Connect backend, docs at:
    https://docs.globus.org/api/auth
    http://globus-integration-examples.readthedocs.io
"""

from social_core.backends.open_id_connect import OpenIdConnectAuth


class GlobusOpenIdConnect(OpenIdConnectAuth):
    name = "globus"
    OIDC_ENDPOINT = "https://auth.globus.org"
    JWT_ALGORITHMS = ["RS256", "RS512"]
    EXTRA_DATA = [
        ("expires_in", "expires_in", True),
        ("refresh_token", "refresh_token", True),
        ("id_token", "id_token", True),
        ("other_tokens", "other_tokens", True),
    ]

    def get_user_details(self, response):
        username_key = self.setting("USERNAME_KEY", default=self.USERNAME_KEY)
        name = response.get("name") or ""
        fullname, first_name, last_name = self.get_user_names(name)
        return {
            "username": response.get(username_key),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class AppsfuelOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.appsfuel.AppsfuelOAuth2"
    user_data_url = "https://api.appsfuel.com/v1/live/user"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "user_id": "123",
            "display_name": "Foo Bar",
            "email": "foobar@example.com",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()


class AppsfuelOAuth2SandboxTest(AppsfuelOAuth2Test):
    backend_path = "social_core.backends.appsfuel.AppsfuelOAuth2Sandbox"
    user_data_url = "https://api.appsfuel.com/v1/sandbox/user"



"""
Okta OAuth2 and OpenIdConnect:
    https://python-social-auth.readthedocs.io/en/latest/backends/okta.html
"""

from typing import Any

from .okta import OktaOAuth2
from .open_id_connect import OpenIdConnectAuth


class OktaOpenIdConnect(OktaOAuth2, OpenIdConnectAuth):
    """Okta OpenID-Connect authentication backend"""

    name = "okta-openidconnect"
    REDIRECT_STATE = False
    RESPONSE_TYPE = "code"

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        return self.validate_userinfo_sub(
            super().user_data(access_token, *args, **kwargs)
        )



"""
Copyright (c) 2015 Microsoft Open Technologies, Inc.

All rights reserved.

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Azure AD OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/azuread.html
"""

from __future__ import annotations

import os
import time
from pathlib import Path
from typing import Any, cast

import jwt

from social_core.exceptions import AuthMissingParameter, AuthTokenError
from social_core.utils import cache

from .oauth import BaseOAuth2


class AzureADOAuth2(BaseOAuth2):
    name = "azuread-oauth2"
    SCOPE_SEPARATOR = " "
    BASE_URL = "https://{authority_host}/{tenant_id}"
    AUTHORIZATION_URL = "{base_url}/oauth2/authorize"
    ACCESS_TOKEN_URL = "{base_url}/oauth2/token"
    OPENID_CONFIGURATION_URL = "{base_url}/.well-known/openid-configuration"
    REDIRECT_STATE = False
    DEFAULT_SCOPE = ["openid", "profile", "user_impersonation", "email"]
    JWT_ALGORITHMS = ["RS256"]
    EXTRA_DATA = [
        ("access_token", "access_token"),
        ("id_token", "id_token"),
        ("refresh_token", "refresh_token"),
        ("expires_in", "expires_in"),
        ("expires_on", "expires_on"),
        ("not_before", "not_before"),
        ("given_name", "first_name"),
        ("family_name", "last_name"),
        ("token_type", "token_type"),
    ]

    @property
    def authority_host(self):
        return self.setting("AUTHORITY_HOST", "login.microsoftonline.com")

    @property
    def tenant_id(self) -> str:
        return "common"

    @property
    def base_url(self) -> str:
        return self.BASE_URL.format(
            authority_host=self.authority_host, tenant_id=self.tenant_id
        )

    def get_authorization_url_format(self) -> dict[str, str]:
        return {"base_url": self.base_url}

    def get_access_token_url_format(self) -> dict[str, str]:
        return {"base_url": self.base_url}

    def openid_configuration_url(self) -> str:
        url = cast(
            "str",
            self.setting("OPENID_CONFIGURATION_URL", self.OPENID_CONFIGURATION_URL),
        )
        return url.format(
            authority_host=self.authority_host,
            base_url=self.base_url,
            tenant_id=self.tenant_id,
        )

    @cache(ttl=86400)
    def get_openid_configuration(self, url: str) -> dict[str, Any]:
        return self.get_json(url)

    def openid_configuration(self) -> dict[str, Any]:
        configuration = self.get_openid_configuration(self.openid_configuration_url())
        return cast("dict[str, Any]", configuration)

    def jwks_uri(self) -> str:
        uri = self.setting("JWKS_URI") or self.openid_configuration().get("jwks_uri")
        if not isinstance(uri, str):
            raise AuthMissingParameter(self, "jwks_uri")
        return uri

    @cache(ttl=86400)
    def get_jwks_keys_for_uri(self, uri: str) -> list[dict[str, Any]]:
        jwks = self.get_json(uri)
        keys = jwks.get("keys")
        if not isinstance(keys, list):
            raise AuthMissingParameter(self, "keys")
        return cast("list[dict[str, Any]]", keys)

    def get_jwks_keys(self) -> list[dict[str, Any]]:
        return self.get_jwks_keys_for_uri(self.jwks_uri())

    def get_user_id(self, details, response):
        """Use upn as unique id"""
        upn = response.get("upn")
        if upn is None:
            raise AuthMissingParameter(self, "upn")
        return upn

    def get_user_details(self, response):
        """Return user details from Azure AD account"""
        fullname, first_name, last_name = (
            response.get("name", ""),
            response.get("given_name", ""),
            response.get("family_name", ""),
        )
        return {
            "username": fullname,
            "email": response.get("email", response.get("upn")),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        response = kwargs.get("response")
        if response and response.get("id_token"):
            id_token = response.get("id_token")
        else:
            id_token = access_token

        return self.validate_and_return_id_token(id_token)

    def get_unverified_claims(self, id_token: str) -> dict[str, Any]:
        try:
            return jwt.decode(
                id_token,
                options={
                    "verify_signature": False,
                    "verify_aud": False,
                    "verify_exp": False,
                    "verify_iat": False,
                    "verify_iss": False,
                    "verify_nbf": False,
                },
            )
        except jwt.PyJWTError as error:
            raise AuthTokenError(self, error) from error

    def resolve_tenant_issuer(
        self, issuer: str, claims: dict[str, Any], parameter: str
    ) -> str:
        if "{tenantid}" not in issuer and "{tenantId}" not in issuer:
            return issuer

        tenant_id = claims.get("tid")
        if not isinstance(tenant_id, str) or not tenant_id:
            raise AuthMissingParameter(self, parameter)
        return issuer.replace("{tenantid}", tenant_id).replace("{tenantId}", tenant_id)

    def get_id_token_issuer(self, claims: dict[str, Any]) -> str:
        issuer = self.openid_configuration().get("issuer")
        if not isinstance(issuer, str):
            raise AuthMissingParameter(self, "issuer")
        return self.resolve_tenant_issuer(issuer, claims, "tid")

    def get_id_token_key(self, id_token: str) -> dict[str, Any]:
        try:
            header = jwt.get_unverified_header(id_token)
        except jwt.PyJWTError as error:
            raise AuthTokenError(self, error) from error

        key_id = header.get("kid")
        if not key_id:
            raise AuthMissingParameter(self, "kid")

        for key in self.get_jwks_keys():
            if key.get("kid") == key_id:
                return key

        cast("Any", self.get_jwks_keys_for_uri).invalidate()
        for key in self.get_jwks_keys():
            if key.get("kid") == key_id:
                return key
        raise AuthTokenError(self, "Signature key not found")

    def validate_key_issuer(self, key: dict[str, Any], claims: dict[str, Any]) -> None:
        key_issuer = key.get("issuer")
        if not isinstance(key_issuer, str):
            return

        expected_issuer = self.resolve_tenant_issuer(key_issuer, claims, "tid")
        if expected_issuer != claims.get("iss"):
            raise AuthTokenError(self, "Token issuer does not match signing key issuer")

    def get_jwt_algorithms(self) -> list[str]:
        return cast("list[str]", self.setting("JWT_ALGORITHMS", self.JWT_ALGORITHMS))

    def validate_and_return_id_token(self, id_token: str) -> dict[str, Any]:
        unverified_claims = self.get_unverified_claims(id_token)
        key = self.get_id_token_key(id_token)
        self.validate_key_issuer(key, unverified_claims)

        if "alg" not in key:
            key = {**key, "alg": self.get_jwt_algorithms()[0]}

        try:
            return jwt.decode(
                id_token,
                key=jwt.PyJWK(key).key,
                algorithms=self.get_jwt_algorithms(),
                audience=self.setting("KEY"),
                issuer=self.get_id_token_issuer(unverified_claims),
                leeway=cast("int", self.setting("JWT_LEEWAY", default=0)),
            )
        except jwt.PyJWTError as error:
            raise AuthTokenError(self, error) from error

    def auth_extra_arguments(self):
        """Return extra arguments needed on auth process."""
        extra_arguments = super().auth_extra_arguments()
        resource = self.setting("RESOURCE")
        if resource:
            extra_arguments.update({"resource": resource})
        return extra_arguments

    def extra_data(
        self,
        user,
        uid: str,
        response: dict[str, Any],
        details: dict[str, Any],
        pipeline_kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        """Return access_token and extra defined names to store in
        extra_data field"""
        data = super().extra_data(user, uid, response, details, pipeline_kwargs)
        data["resource"] = self.setting("RESOURCE")
        return data

    def refresh_token_params(self, token, *args, **kwargs):
        client_secret = cast("str | None", self.setting("SECRET"))
        params = {
            "client_id": self.setting("KEY"),
            "refresh_token": token,
            "grant_type": "refresh_token",
            "resource": self.setting("RESOURCE"),
        }

        if client_secret:
            params["client_secret"] = client_secret
            return params

        assertion = self.client_assertion(required=True)
        params.update(
            {
                "client_assertion_type": self.client_assertion_type(),
                "client_assertion": assertion,
            }
        )

        return params

    def get_auth_token(self, user_id):
        """Return the access token for the given user, after ensuring that it
        has not expired, or refreshing it if so."""
        user = self.get_user(user_id=user_id)
        access_token = user.social_user.access_token
        expires_on = user.social_user.extra_data["expires_on"]
        if expires_on <= int(time.time()):
            new_token_response = self.refresh_token(token=access_token)
            access_token = new_token_response["access_token"]
        return access_token

    def auth_complete_params(self, state=None):
        params = super().auth_complete_params(state)
        if params.get("client_secret"):
            return params

        assertion = self.client_assertion(required=True)
        params.update(
            {
                "client_assertion_type": self.client_assertion_type(),
                "client_assertion": assertion,
            }
        )
        return params

    def client_assertion(self, *, required: bool = False) -> str | None:
        if cast("str | None", self.setting("SECRET")):
            return None

        assertion = self.setting("CLIENT_ASSERTION")
        if assertion:
            return assertion

        token_path = (
            os.environ.get(
                "OAUTH2_FEDERATED_TOKEN_FILE"
            )  # supports OAUTH2_ naming convention
            or os.environ.get("AZURE_FEDERATED_TOKEN_FILE")  # canonical name
            or self.setting("FEDERATED_TOKEN_FILE")
        )
        if not token_path:
            if required:
                raise AuthMissingParameter(self, "client_assertion")
            return None

        try:
            assertion = Path(token_path).read_text(encoding="utf-8").strip()
        except OSError as error:
            if required:
                raise AuthMissingParameter(self, "client_assertion") from error
            return None

        if not assertion:
            if required:
                raise AuthMissingParameter(self, "client_assertion")
            return None

        return assertion

    def client_assertion_type(self) -> str:
        return cast(
            "str",
            self.setting(
                "CLIENT_ASSERTION_TYPE",
                "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
            ),
        )


class AzureADOAuth2V2(AzureADOAuth2):
    """Version of the AzureADOAuth2 backend that uses the v2.0 API endpoints,
    supporting users with personal Microsoft accounts, if the app settings
    allow them."""

    name = "azuread-oauth2-v2"
    AUTHORIZATION_URL = "{base_url}/oauth2/v2.0/authorize"
    ACCESS_TOKEN_URL = "{base_url}/oauth2/v2.0/token"
    OPENID_CONFIGURATION_URL = "{base_url}/v2.0/.well-known/openid-configuration"
    DEFAULT_SCOPE = ["User.Read profile openid email"]



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2StateTestMixin, OAuth2Test


class SurveyMonkeyOAuth2Test(OAuth2Test, OAuth2StateTestMixin, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.surveymonkey.SurveyMonkeyOAuth2"
    user_data_url = "https://api.surveymonkey.com/v3/users/me"
    expected_username = "ada"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "access_url": "https://api.surveymonkey.com",
            "token_type": "bearer",
        }
    )
    user_data_body = json.dumps(
        {
            "id": "123",
            "username": "ada",
            "email": "ada@example.com",
            "first_name": "Ada",
            "last_name": "Lovelace",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



from __future__ import annotations

from social_core.backends.base import BaseAuth
from social_core.tests.models import TestStorage
from social_core.tests.strategy import TestStrategy


class ExampleAuth(BaseAuth):
    name = "example"


def get_backend(settings, request_data=None):
    strategy = TestStrategy(TestStorage)
    backend = ExampleAuth(strategy)
    strategy.set_settings(settings)
    if request_data is not None:
        strategy.set_request_data(request_data, backend)
    return backend


def test_auth_extra_arguments_ignore_matching_request_parameters_by_default() -> None:
    backend = get_backend(
        {
            "SOCIAL_AUTH_EXAMPLE_AUTH_EXTRA_ARGUMENTS": {
                "audience": "configured-audience",
                "resource": "configured-resource",
            }
        },
        {
            "audience": "request-audience",
            "resource": "request-resource",
        },
    )

    assert backend.auth_extra_arguments() == {
        "audience": "configured-audience",
        "resource": "configured-resource",
    }


def test_auth_extra_arguments_allowlisted_key_can_use_request_parameter() -> None:
    backend = get_backend(
        {
            "SOCIAL_AUTH_EXAMPLE_AUTH_EXTRA_ARGUMENTS": {
                "audience": "configured-audience",
                "prompt": "login",
                "resource": "configured-resource",
            },
            "SOCIAL_AUTH_EXAMPLE_AUTH_EXTRA_ARGUMENTS_OVERRIDE_ALLOWLIST": ["prompt"],
        },
        {
            "audience": "request-audience",
            "prompt": "select_account",
            "resource": "request-resource",
        },
    )

    assert backend.auth_extra_arguments() == {
        "audience": "configured-audience",
        "prompt": "select_account",
        "resource": "configured-resource",
    }


def test_auth_extra_arguments_request_only_keys_are_not_added() -> None:
    backend = get_backend(
        {
            "SOCIAL_AUTH_EXAMPLE_AUTH_EXTRA_ARGUMENTS": {"prompt": "login"},
            "SOCIAL_AUTH_EXAMPLE_AUTH_EXTRA_ARGUMENTS_OVERRIDE_ALLOWLIST": [
                "prompt",
                "audience",
            ],
        },
        {
            "audience": "request-audience",
            "prompt": "select_account",
        },
    )

    assert backend.auth_extra_arguments() == {"prompt": "select_account"}



import json
import time

import responses

from social_core.utils import parse_qs

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class MicrosoftOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.microsoft.MicrosoftOAuth2"
    user_data_url = "https://graph.microsoft.com/v1.0/me"
    expected_username = "foobar"
    user_data_body = json.dumps(
        {
            "displayName": "foo bar",
            "givenName": "foobar",
            "jobTitle": "Auditor",
            "mail": "foobar@foobar.com",
            "mobilePhone": None,
            "officeLocation": "12/1110",
            "preferredLanguage": "en-US",
            "surname": "Bowen",
            "userPrincipalName": "foobar",
            "id": "48d31887-5fad-4d73-a9f5-3c356e68a038",
        }
    )
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "id_token": "",
            "expires_in": 3600,
            "expires_on": 1423650396,
            "not_before": 1423646496,
            "refresh_token": "foobar-refresh-token",
        }
    )
    refresh_token_body = json.dumps(
        {
            "access_token": "foobar-new-token",
            "token_type": "bearer",
            "expires_in": 3600,
            "refresh_token": "foobar-new-refresh-token",
            "scope": "identity",
        }
    )

    def test_login(self) -> None:
        user = self.do_login()
        social = user.social[0]
        self.assertEqual(social.extra_data["refresh_token"], "foobar-refresh-token")
        self.assertEqual(social.extra_data["expires_in"], 3600)

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()

    def test_refresh_token(self) -> None:
        _user, social = self.do_refresh_token()
        self.assertEqual(social.extra_data["access_token"], "foobar-new-token")
        self.assertEqual(social.extra_data["refresh_token"], "foobar-new-refresh-token")
        self.assertEqual(
            parse_qs(responses.calls[-1].request.body)["refresh_token"],
            "foobar-refresh-token",
        )

    def test_get_access_token_refreshes_expired_token(self) -> None:
        user = self.do_login()
        responses.add(
            self._method(self.backend.REFRESH_TOKEN_METHOD),
            self.backend.refresh_token_url(),
            status=200,
            body=self.refresh_token_body,
        )
        social = user.social[0]
        social.extra_data["auth_time"] = int(time.time()) - 7200
        self.assertEqual(social.get_access_token(self.strategy), "foobar-new-token")
        self.assertEqual(social.extra_data["refresh_token"], "foobar-new-refresh-token")

    def test_get_auth_token_refreshes_expired_token(self) -> None:
        user = self.do_login()
        responses.add(
            self._method(self.backend.REFRESH_TOKEN_METHOD),
            self.backend.refresh_token_url(),
            status=200,
            body=self.refresh_token_body,
        )
        user.social_user.extra_data["auth_time"] = int(time.time()) - 7200
        self.assertEqual(self.backend.get_auth_token(user.id), "foobar-new-token")



import json
from typing import Any
from unittest import TestCase

import responses

from social_core.exceptions import AuthFailed

from .oauth import BaseAuthUrlTestMixin, OAuth2Test

CAPTURE_GITHUB_EMAILS_PIPELINE = (
    "social_core.tests.backends.test_github.capture_github_emails"
)


def capture_github_emails(strategy, response, *args, **kwargs) -> None:
    strategy.session_set("github_emails", response.get("emails"))


class TestCaptureGithubEmails(TestCase):
    class DummyStrategy:
        def __init__(self) -> None:
            self.data: dict[str, Any] = {}

        def session_set(self, key, value) -> None:
            self.data[key] = value

    def test_capture_github_emails_missing_emails_key(self) -> None:
        strategy = self.DummyStrategy()

        capture_github_emails(strategy, {})

        assert "github_emails" in strategy.data
        assert strategy.data["github_emails"] is None

    def test_capture_github_emails_response_none_raises_attribute_error(self) -> None:
        strategy = self.DummyStrategy()

        with self.assertRaises(AttributeError):
            capture_github_emails(strategy, None)


class GithubOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.github.GithubOAuth2"
    user_data_url = "https://api.github.com/user"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "expires_in": 28800,
            "refresh_token": "foobar-refresh-token",
        }
    )
    refresh_token_body = json.dumps(
        {
            "access_token": "foobar-new-token",
            "token_type": "bearer",
            "expires_in": 28800,
            "refresh_token": "foobar-new-refresh-token",
            "refresh_token_expires_in": 15897600,
            "scope": "",
        }
    )
    user_data_body = json.dumps(
        {
            "login": "foobar",
            "id": 1,
            "avatar_url": "https://github.com/images/error/foobar_happy.gif",
            "gravatar_id": "somehexcode",
            "url": "https://api.github.com/users/foobar",
            "name": "monalisa foobar",
            "company": "GitHub",
            "blog": "https://github.com/blog",
            "location": "San Francisco",
            "email": "foo@bar.com",
            "hireable": False,
            "bio": "There once was...",
            "public_repos": 2,
            "public_gists": 1,
            "followers": 20,
            "following": 0,
            "html_url": "https://github.com/foobar",
            "created_at": "2008-01-14T04:33:35Z",
            "type": "User",
            "total_private_repos": 100,
            "owned_private_repos": 100,
            "private_gists": 81,
            "disk_usage": 10000,
            "collaborators": 8,
            "plan": {
                "name": "Medium",
                "space": 400,
                "collaborators": 10,
                "private_repos": 20,
            },
        }
    )

    def do_login(self):
        user = super().do_login()
        self.assertTrue(user.social)
        social = user.social[0]

        self.assertIsNotNone(social.extra_data["expires_in"])
        self.assertIsNotNone(social.extra_data["refresh_token"])

        return user

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()

    def test_refresh_token(self) -> None:
        _user, social = self.do_refresh_token()
        self.assertEqual(social.extra_data["access_token"], "foobar-new-token")


class GithubOAuth2NoEmailTest(GithubOAuth2Test):
    emails_url = "https://api.github.com/user/emails"
    user_data_body = json.dumps(
        {
            "login": "foobar",
            "id": 1,
            "avatar_url": "https://github.com/images/error/foobar_happy.gif",
            "gravatar_id": "somehexcode",
            "url": "https://api.github.com/users/foobar",
            "name": "monalisa foobar",
            "company": "GitHub",
            "blog": "https://github.com/blog",
            "location": "San Francisco",
            "email": "",
            "hireable": False,
            "bio": "There once was...",
            "public_repos": 2,
            "public_gists": 1,
            "followers": 20,
            "following": 0,
            "html_url": "https://github.com/foobar",
            "created_at": "2008-01-14T04:33:35Z",
            "type": "User",
            "total_private_repos": 100,
            "owned_private_repos": 100,
            "private_gists": 81,
            "disk_usage": 10000,
            "collaborators": 8,
            "plan": {
                "name": "Medium",
                "space": 400,
                "collaborators": 10,
                "private_repos": 20,
            },
        }
    )

    def add_emails_response(
        self, emails: list[dict[str, str | bool]], status: int = 200
    ) -> None:
        responses.add(
            responses.GET,
            self.emails_url,
            status=status,
            body=json.dumps(emails),
            content_type="application/json",
        )

    def capture_emails_in_pipeline(self) -> None:
        pipeline = self.strategy.get_pipeline(self.backend)
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_PIPELINE": (
                    pipeline[0],
                    CAPTURE_GITHUB_EMAILS_PIPELINE,
                    *pipeline[1:],
                )
            }
        )

    def test_login_email_denied(self) -> None:
        self.add_emails_response([], status=403)
        self.do_login()

    def test_login_with_empty_email_list(self) -> None:
        self.add_emails_response([])
        user = self.do_login()
        self.assertNotIn("emails", user.social[0].extra_data)

    def test_login_next_format(self) -> None:
        self.add_emails_response([{"email": "foo@bar.com"}])
        user = self.do_login()
        self.assertEqual(user.email, "foo@bar.com")

    def test_login(self) -> None:
        emails: list[dict[str, str | bool]] = [
            {"email": "secondary@example.com", "primary": False},
            {"email": "foo@bar.com", "primary": True},
        ]
        self.add_emails_response(emails)
        self.capture_emails_in_pipeline()

        user = self.do_login()

        self.assertEqual(self.strategy.session_get("github_emails"), emails)
        self.assertEqual(user.email, "foo@bar.com")
        self.assertNotIn("emails", user.social[0].extra_data)

    def test_partial_pipeline(self) -> None:
        self.add_emails_response([{"email": "foo@bar.com"}])
        self.do_partial_pipeline()

    def test_refresh_token(self) -> None:
        self.add_emails_response([{"email": "foo@bar.com"}])
        self.do_refresh_token()


class GithubOrganizationOAuth2Test(GithubOAuth2Test):
    backend_path = "social_core.backends.github.GithubOrganizationOAuth2"

    def auth_handlers(self, start_url):
        url = "https://api.github.com/orgs/foobar/members/foobar"
        responses.add(responses.GET, url, status=204, body="")
        return super().auth_handlers(start_url)

    def test_login(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ORG_NAME": "foobar"})
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ORG_NAME": "foobar"})
        self.do_partial_pipeline()

    def test_refresh_token(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ORG_NAME": "foobar"})
        self.do_refresh_token()


class GithubOrganizationOAuth2FailTest(GithubOAuth2Test):
    backend_path = "social_core.backends.github.GithubOrganizationOAuth2"

    def auth_handlers(self, start_url):
        url = "https://api.github.com/orgs/foobar/members/foobar"
        responses.add(
            responses.GET,
            url,
            status=404,
            body='{"message": "Not Found"}',
            content_type="application/json",
        )
        return super().auth_handlers(start_url)

    def test_login(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ORG_NAME": "foobar"})
        with self.assertRaises(AuthFailed):
            self.do_login()

    def test_partial_pipeline(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ORG_NAME": "foobar"})
        with self.assertRaises(AuthFailed):
            self.do_partial_pipeline()

    def test_refresh_token(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ORG_NAME": "foobar"})
        with self.assertRaises(AuthFailed):
            self.do_refresh_token()


class GithubTeamOAuth2Test(GithubOAuth2Test):
    backend_path = "social_core.backends.github.GithubTeamOAuth2"

    def auth_handlers(self, start_url):
        url = "https://api.github.com/teams/123/members/foobar"
        responses.add(responses.GET, url, status=204, body="")
        return super().auth_handlers(start_url)

    def test_login(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_TEAM_ID": "123"})
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_TEAM_ID": "123"})
        self.do_partial_pipeline()

    def test_refresh_token(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_TEAM_ID": "123"})
        self.do_refresh_token()


class GithubTeamOAuth2FailTest(GithubOAuth2Test):
    backend_path = "social_core.backends.github.GithubTeamOAuth2"

    def auth_handlers(self, start_url):
        url = "https://api.github.com/teams/123/members/foobar"
        responses.add(
            responses.GET,
            url,
            status=404,
            body='{"message": "Not Found"}',
            content_type="application/json",
        )
        return super().auth_handlers(start_url)

    def test_login(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_TEAM_ID": "123"})
        with self.assertRaises(AuthFailed):
            self.do_login()

    def test_partial_pipeline(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_TEAM_ID": "123"})
        with self.assertRaises(AuthFailed):
            self.do_partial_pipeline()

    def test_refresh_token(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_TEAM_ID": "123"})
        with self.assertRaises(AuthFailed):
            self.do_refresh_token()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2StateTestMixin, OAuth2Test


class MailChimpOAuth2Test(OAuth2Test, OAuth2StateTestMixin, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.mailchimp.MailChimpOAuth2"
    user_data_url = "https://login.mailchimp.com/oauth2/metadata"
    expected_username = "ada"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "accountname": "Example",
            "api_endpoint": "https://us1.api.mailchimp.com",
            "login": {"login_name": "ada", "email": "ada@example.com"},
            "role": "owner",
            "user_id": "123",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import json

import responses

from .base import BaseBackendTest


class LastFmAuthTest(BaseBackendTest):
    backend_path = "social_core.backends.lastfm.LastFmAuth"
    expected_username = "foobar"

    def extra_settings(self) -> dict[str, str | list[str]]:
        return {
            "SOCIAL_AUTH_LASTFM_KEY": "a-key",
            "SOCIAL_AUTH_LASTFM_SECRET": "a-secret-key",
        }

    def do_start(self):
        start_url = self.backend.start().url
        self.assertEqual(start_url, "https://www.last.fm/api/auth/?api_key=a-key")
        self.strategy.set_request_data({"token": "foobar"}, self.backend)
        responses.add(
            responses.POST,
            "https://ws.audioscrobbler.com/2.0/",
            body=json.dumps({"session": {"name": "foobar", "key": "session-key"}}),
            content_type="application/json",
        )
        return self.backend.complete()

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2StateTestMixin, OAuth2Test


class UntappdOAuth2Test(OAuth2Test, OAuth2StateTestMixin, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.untappd.UntappdOAuth2"
    user_data_url = "https://api.untappd.com/v4/user/info/"
    expected_username = "ada"
    access_token_body = json.dumps(
        {"meta": {"http_code": 200}, "response": {"access_token": "foobar"}}
    )
    user_data_body = json.dumps(
        {
            "meta": {"http_code": 200},
            "response": {
                "user": {
                    "id": "123",
                    "user_name": "ada",
                    "first_name": "Ada",
                    "last_name": "Lovelace",
                    "settings": {"email_address": "ada@example.com"},
                }
            },
        }
    )

    def test_auth_params_include_redirect_url_and_state(self) -> None:
        params = self.backend.auth_params("test-state")

        self.assertEqual(params["redirect_url"], self.backend.get_redirect_uri())
        self.assertEqual(params["state"], "test-state")

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Copyright (c) 2015 Microsoft Open Technologies, Inc.

All rights reserved.

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import base64
import json
import os
import tempfile
from time import time
from typing import TYPE_CHECKING, cast
from unittest import TestCase
from unittest.mock import patch
from urllib.parse import parse_qs

import jwt
import responses
from jwt.algorithms import RSAAlgorithm

from social_core.exceptions import AuthMissingParameter, AuthTokenError

from .oauth import BaseAuthUrlTestMixin, OAuth2Test
from .test_azuread_b2c import RSA_PRIVATE_JWT_KEY, RSA_PUBLIC_JWT_KEY

if TYPE_CHECKING:
    from typing import Any

    from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey


class AzureADOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    AUTH_KEY = "a-key"
    AUTH_TIME = int(time())
    EXPIRES_IN = 3600
    EXPIRES_ON = AUTH_TIME + EXPIRES_IN
    JWKS_URL = "https://login.microsoftonline.com/common/discovery/keys"
    ISSUER = "https://sts.windows.net/727406ac-7068-48fa-92b9-c2d67211bc50/"
    ISSUER_TEMPLATE = "https://sts.windows.net/{tenantid}/"
    KEY_ISSUER = ISSUER_TEMPLATE
    TENANT_ID = "727406ac-7068-48fa-92b9-c2d67211bc50"
    TOKEN_VERSION = "1.0"

    backend_path = "social_core.backends.azuread.AzureADOAuth2"
    user_data_url = "https://graph.windows.net/me"
    expected_username = "foobar"
    access_token_body = ""
    refresh_token_body = json.dumps(
        {
            "access_token": "foobar-new-token",
            "token_type": "bearer",
            "expires_in": 3600,
            "refresh_token": "foobar-new-refresh-token",
            "scope": "identity",
        }
    )

    def setUp(self) -> None:
        super().setUp()
        self.access_token_body = self.build_access_token_body()
        responses.add(
            responses.GET,
            self.backend.openid_configuration_url(),
            body=json.dumps(
                {
                    "issuer": self.ISSUER_TEMPLATE,
                    "jwks_uri": self.JWKS_URL,
                    "id_token_signing_alg_values_supported": ["RS256"],
                }
            ),
            content_type="application/json",
        )
        responses.add(
            responses.GET,
            self.JWKS_URL,
            body=json.dumps(
                {"keys": [{**RSA_PUBLIC_JWT_KEY, "issuer": self.KEY_ISSUER}]}
            ),
            content_type="application/json",
        )

    def build_id_token(self, **overrides) -> str:
        payload = {
            "aud": self.AUTH_KEY,
            "exp": self.EXPIRES_ON,
            "family_name": "bar",
            "given_name": "foo",
            "iat": self.AUTH_TIME,
            "iss": self.ISSUER,
            "name": "foo bar",
            "nbf": self.AUTH_TIME,
            "oid": "7f8e1969-8b81-438c-8d4e-ad6f562b28bb",
            "preferred_username": "foobar@test.onmicrosoft.com",
            "sub": "qU8xks9mHqnV6Q34zh7SAZocih9EzrrI9mpVXOiBVA8",
            "tid": self.TENANT_ID,
            "upn": "foobar@test.onmicrosoft.com",
            "ver": self.TOKEN_VERSION,
        }
        payload.update(overrides)
        return jwt.encode(
            payload,
            key=cast(
                "RSAPrivateKey",
                RSAAlgorithm.from_jwk(json.dumps(RSA_PRIVATE_JWT_KEY)),
            ),
            algorithm="RS256",
            headers={"kid": RSA_PRIVATE_JWT_KEY["kid"]},
        )

    def build_access_token_body(self, **id_token_overrides) -> str:
        return json.dumps(
            {
                "access_token": "foobar",
                "token_type": "bearer",
                "id_token": self.build_id_token(**id_token_overrides),
                "expires_in": self.EXPIRES_IN,
                "expires_on": self.EXPIRES_ON,
                "not_before": self.AUTH_TIME,
            }
        )

    def tamper_id_token(self, id_token: str, **overrides) -> str:
        header, _payload, signature = id_token.split(".")
        payload = jwt.decode(
            id_token,
            options={"verify_signature": False, "verify_aud": False},
        )
        payload.update(overrides)
        data = json.dumps(payload, separators=(",", ":")).encode()
        tampered_payload = base64.urlsafe_b64encode(data).rstrip(b"=").decode()
        return f"{header}.{tampered_payload}.{signature}"

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()

    def test_refresh_token(self) -> None:
        _user, social = self.do_refresh_token()
        self.assertEqual(social.extra_data["access_token"], "foobar-new-token")

    def test_login_rejects_invalid_id_token_signature(self) -> None:
        id_token = self.build_id_token()
        self.access_token_body = json.dumps(
            {
                "access_token": "foobar",
                "token_type": "bearer",
                "id_token": self.tamper_id_token(id_token, upn="attacker@example.com"),
            }
        )

        with self.assertRaises(AuthTokenError):
            self.do_start()

    def test_login_rejects_wrong_id_token_audience(self) -> None:
        self.access_token_body = self.build_access_token_body(aud="other-app")

        with self.assertRaises(AuthTokenError):
            self.do_start()

    def test_login_rejects_wrong_id_token_issuer(self) -> None:
        self.access_token_body = self.build_access_token_body(
            iss="https://sts.windows.net/00000000-0000-0000-0000-000000000000/"
        )

        with self.assertRaises(AuthTokenError):
            self.do_start()

    def test_login_rejects_invalid_id_token_tenant_id(self) -> None:
        self.access_token_body = self.build_access_token_body(tid=None)

        with self.assertRaises(AuthMissingParameter):
            self.do_start()

    def test_openid_configuration_and_jwks_cache_shared_by_url(self) -> None:
        cast("Any", self.backend.get_openid_configuration).invalidate()
        cast("Any", self.backend.get_jwks_keys_for_uri).invalidate()

        other_backend = self.backend.__class__(
            self.strategy, redirect_uri=self.complete_url
        )

        self.backend.openid_configuration()
        other_backend.openid_configuration()
        self.backend.get_jwks_keys()
        other_backend.get_jwks_keys()

        self.assertEqual(
            self.response_call_count(self.backend.openid_configuration_url()), 1
        )
        self.assertEqual(self.response_call_count(self.JWKS_URL), 1)

    def response_call_count(self, url: str) -> int:
        return sum(1 for call in responses.calls if call.request.url == url)


class AzureADOAuth2V2Test(AzureADOAuth2Test):
    backend_path = "social_core.backends.azuread.AzureADOAuth2V2"
    ISSUER = (
        "https://login.microsoftonline.com/727406ac-7068-48fa-92b9-c2d67211bc50/v2.0"
    )
    ISSUER_TEMPLATE = "https://login.microsoftonline.com/{tenantid}/v2.0"
    KEY_ISSUER = ISSUER_TEMPLATE
    JWKS_URL = "https://login.microsoftonline.com/common/discovery/v2.0/keys"
    TOKEN_VERSION = "2.0"


class AzureADTenantOAuth2Test(AzureADOAuth2Test):
    backend_path = "social_core.backends.azuread_tenant.AzureADTenantOAuth2"
    JWKS_URL = (
        "https://login.microsoftonline.com/"
        f"{AzureADOAuth2Test.TENANT_ID}/discovery/keys?appid=a-key"
    )

    def extra_settings(self):
        settings = super().extra_settings()
        assert self.name, "Name must be set in subclasses"
        settings[f"SOCIAL_AUTH_{self.name}_TENANT_ID"] = self.TENANT_ID
        return settings

    def issuer_for_tenant(self, tenant_id: str) -> str:
        return self.ISSUER.replace(self.TENANT_ID, tenant_id)

    def test_login_rejects_wrong_configured_tenant_id(self) -> None:
        other_tenant_id = "00000000-0000-0000-0000-000000000000"
        self.access_token_body = self.build_access_token_body(
            iss=self.issuer_for_tenant(other_tenant_id),
            tid=other_tenant_id,
        )

        with self.assertRaises(AuthTokenError):
            self.do_start()


class AzureADV2TenantOAuth2Test(AzureADTenantOAuth2Test):
    backend_path = "social_core.backends.azuread_tenant.AzureADV2TenantOAuth2"
    ISSUER = (
        "https://login.microsoftonline.com/727406ac-7068-48fa-92b9-c2d67211bc50/v2.0"
    )
    ISSUER_TEMPLATE = "https://login.microsoftonline.com/{tenantid}/v2.0"
    KEY_ISSUER = ISSUER_TEMPLATE
    JWKS_URL = (
        "https://login.microsoftonline.com/"
        f"{AzureADOAuth2Test.TENANT_ID}/discovery/v2.0/keys?appid=a-key"
    )
    TOKEN_VERSION = "2.0"


class AzureADOAuth2TokenRequestBodyMixin(TestCase):
    def _token_request_body(self, url_prefix: str) -> dict[str, list[str]]:
        matches = [
            call.request
            for call in responses.calls
            if cast("str", call.request.url).startswith(url_prefix)
        ]
        self.assertGreaterEqual(
            len(matches),
            1,
            f"expected at least one token request for {url_prefix}, found {len(matches)}",
        )
        request = matches[-1]
        body = request.body or ""
        if isinstance(body, bytes):
            body = body.decode()
        return parse_qs(body)


class AzureADOAuth2FederatedIdentityCredentialTest(
    AzureADOAuth2TokenRequestBodyMixin, AzureADOAuth2Test
):
    def extra_settings(self):
        settings = super().extra_settings()
        settings.pop("SOCIAL_AUTH_AZUREAD_OAUTH2_SECRET", None)
        settings["SOCIAL_AUTH_AZUREAD_OAUTH2_CLIENT_ASSERTION"] = "fic-assertion"
        return settings

    def test_login_uses_client_assertion(self) -> None:
        self.do_login()
        body = self._token_request_body(self.backend.access_token_url())
        self.assertIn("client_assertion", body)
        self.assertEqual(body["client_assertion"], ["fic-assertion"])
        self.assertEqual(
            body.get("client_assertion_type"),
            ["urn:ietf:params:oauth:client-assertion-type:jwt-bearer"],
        )
        self.assertNotIn("client_secret", body)

    def test_refresh_token_uses_client_assertion(self) -> None:
        self.do_refresh_token()
        body = self._token_request_body(self.backend.refresh_token_url())
        self.assertIn("client_assertion", body)
        self.assertNotIn("client_secret", body)


class AzureADOAuth2FederatedIdentityCredentialFromFileTest(
    AzureADOAuth2TokenRequestBodyMixin, AzureADOAuth2Test
):
    def setUp(self) -> None:
        super().setUp()
        # Default token file for class-level flows; individual tests can override.
        self.token_path = self._write_temp_token("default-assertion")
        patcher = patch.dict(
            os.environ,
            {
                "OAUTH2_FEDERATED_TOKEN_FILE": self.token_path,
                "AZURE_FEDERATED_TOKEN_FILE": self.token_path,
            },
            clear=False,
        )
        patcher.start()
        self.addCleanup(patcher.stop)

    def extra_settings(self):
        settings = super().extra_settings()
        settings.pop("SOCIAL_AUTH_AZUREAD_OAUTH2_SECRET", None)
        settings.pop("SOCIAL_AUTH_AZUREAD_OAUTH2_CLIENT_ASSERTION", None)
        settings.pop("SOCIAL_AUTH_AZUREAD_OAUTH2_FEDERATED_TOKEN_FILE", None)
        return settings

    def _write_temp_token(self, value: str) -> str:
        with tempfile.NamedTemporaryFile("w", delete=False) as handle:
            handle.write(value)
            self.addCleanup(os.remove, handle.name)
            return handle.name

    def test_login_uses_oauth2_env_token_path(self) -> None:
        token_path = self._write_temp_token("env-assertion")
        with patch.dict(
            os.environ, {"OAUTH2_FEDERATED_TOKEN_FILE": token_path}, clear=False
        ):
            self.do_login()

        body = self._token_request_body(self.backend.access_token_url())
        self.assertEqual(body.get("client_assertion"), ["env-assertion"])
        self.assertEqual(
            body.get("client_assertion_type"),
            ["urn:ietf:params:oauth:client-assertion-type:jwt-bearer"],
        )
        self.assertNotIn("client_secret", body)

    def test_login_uses_azure_env_token_path(self) -> None:
        token_path = self._write_temp_token("azure-env-assertion")
        with patch.dict(
            os.environ,
            {
                "OAUTH2_FEDERATED_TOKEN_FILE": "",
                "AZURE_FEDERATED_TOKEN_FILE": token_path,
            },
            clear=False,
        ):
            self.do_login()

        body = self._token_request_body(self.backend.access_token_url())
        self.assertEqual(body.get("client_assertion"), ["azure-env-assertion"])
        self.assertNotIn("client_secret", body)

    def test_missing_env_token_path_raises(self) -> None:
        with (
            patch.dict(
                os.environ,
                {"OAUTH2_FEDERATED_TOKEN_FILE": "/no/such/file"},
                clear=False,
            ),
            self.assertRaises(AuthMissingParameter),
        ):
            self.do_login()

    def test_empty_token_file_raises(self) -> None:
        token_path = self._write_temp_token(" \n\t")
        with (
            patch.dict(
                os.environ,
                {
                    "OAUTH2_FEDERATED_TOKEN_FILE": token_path,
                    "AZURE_FEDERATED_TOKEN_FILE": token_path,
                },
                clear=False,
            ),
            self.assertRaises(AuthMissingParameter),
        ):
            self.do_login()

    def test_empty_token_file_optional_returns_none(self) -> None:
        token_path = self._write_temp_token(" \n\t")
        with patch.dict(
            os.environ,
            {
                "OAUTH2_FEDERATED_TOKEN_FILE": token_path,
                "AZURE_FEDERATED_TOKEN_FILE": token_path,
            },
            clear=False,
        ):
            assertion = self.backend.client_assertion(required=False)

        self.assertIsNone(assertion)


class AzureADOAuth2MissingCredentialsTest(AzureADOAuth2Test):
    def extra_settings(self):
        settings = super().extra_settings()
        settings.pop("SOCIAL_AUTH_AZUREAD_OAUTH2_SECRET", None)
        settings.pop("SOCIAL_AUTH_AZUREAD_OAUTH2_CLIENT_ASSERTION", None)
        return settings

    def test_missing_secret_and_assertion_fails(self) -> None:
        with (
            patch.dict(
                os.environ,
                {
                    "AZURE_FEDERATED_TOKEN_FILE": "",
                    "OAUTH2_FEDERATED_TOKEN_FILE": "",
                },
                clear=False,
            ),
            self.assertRaises(AuthMissingParameter),
        ):
            self.do_login()

    def test_login(self) -> None:
        with self.assertRaises(AuthMissingParameter):
            super().test_login()

    def test_partial_pipeline(self) -> None:
        with self.assertRaises(AuthMissingParameter):
            super().test_partial_pipeline()

    def test_refresh_token(self) -> None:
        with self.assertRaises(AuthMissingParameter):
            super().test_refresh_token()

    def test_login_rejects_invalid_id_token_signature(self) -> None:
        with self.assertRaises(AuthMissingParameter):
            super().test_login_rejects_invalid_id_token_signature()

    def test_login_rejects_wrong_id_token_audience(self) -> None:
        with self.assertRaises(AuthMissingParameter):
            super().test_login_rejects_wrong_id_token_audience()

    def test_login_rejects_wrong_id_token_issuer(self) -> None:
        with self.assertRaises(AuthMissingParameter):
            super().test_login_rejects_wrong_id_token_issuer()
